import paramiko
import os
from time import sleep, time
from os import listdir
from os.path import isfile, join
import timeit
import socket
import pandas as pd
import logging, sys

#python 3
import configparser

# define our config file
from config import Config
f = open('./dashc_4K_DP/config_files/config.cfg')
cfg = Config(f)

# set the root folder location for all files
def root_location(file_location):
    root_folder="dashc_4K_DP"
    fileLocation = str(os.path.join(root_folder+"/"+file_location))
    return fileLocation

paramiko.util.log_to_file(root_location("dashc_player.log"))
# set the logging output of paramiko to warning only
logging.getLogger("paramiko").setLevel(logging.INFO)

# read the bandwidth from the 4G csv file
def readCsvThr(file_path):
    pdf = pd.read_csv(file_path, usecols=["DL_bitrate"])
    time_interval = 1000
    time_bw_array = []
    for index, row in pdf.iterrows():
        time_bw_array.append([time_interval, row["DL_bitrate"]])
    m_time_bw_array = modify_zero_thr2(time_bw_array)
    return m_time_bw_array

# if the bandwidth is very very low, increase to a very smal value to remove issues
def modify_zero_thr3(bw_array):
    for i in range(len(bw_array)):
        if bw_array[i][1] <= 0.5:
            bw_array[i][1] = 1.0/1000
    return bw_array

def modify_zero_thr2(bw_array):
    zero_array = []
    for i in range(len(bw_array)):
        if bw_array[i][1] <= 1:
            zero_array.append(bw_array[i][1])
        else:
            len_z_a = len(zero_array)
            diff = 0
            if len_z_a > 0:
                one_packet_zero_range_kbit = ((1500*8)/1000.0)/len_z_a
                if one_packet_zero_range_kbit < 1:
                    diff = 1 - one_packet_zero_range_kbit
                    one_packet_zero_range_kbit = 1
                for j in range(1, len_z_a+1):
                    bw_array[i-j][1] = one_packet_zero_range_kbit
                sum_val = (len_z_a*one_packet_zero_range_kbit - sum(zero_array)) +diff*len_z_a
                #print sum_val
                zero_array = []
                bw_array[i][1] = bw_array[i][1] - sum_val
                if bw_array[i][1] <= 1:
                    zero_array.append(bw_array[i][1])
    return bw_array

def modify_zero_thr(bw_array):
    diff = 0
    for i in range(0,len(bw_array)):
        if bw_array[i][1] <= 8:
            diff = diff + (8-bw_array[i][1])
            bw_array[i][1] = 8
        elif diff > 0:
            #print "elif"
            if bw_array[i][1] <= diff:
                res = bw_array[i][1] - 8
                bw_array[i][1] = 8
                diff = diff - res
            else:
                #print "else"
                leftovers = bw_array[i][1] - diff
                if leftovers < 8:
                    diff = diff - (bw_array[i][1] - 8)
                    bw_array[i][1] = 8
                else:
                    bw_array[i][1] = leftovers

            diff = 0
    return bw_array

# read the 3G bandwidth log file
def readLog3G(file_log):
    f = open(file_log,'r')
    bw_array = []
    diff = 0
    for lines in f.readlines():
      slpt = float(lines.split()[-1])
      vall = int(8*float(lines.split()[-2])/slpt)
      bw_array.append([slpt,vall])
    f.close()
    m_time_bw_array = modify_zero_thr2(bw_array)
    return m_time_bw_array

# read the bandwidth log file
def readLog(file_log):
    f = open(file_log,'r')
    bw_array = []
    diff = 0
    for lines in f.readlines():
      slpt = float(lines.split()[0])
      vall = float(lines.split()[1])
      bw_array.append([slpt,vall])
    f.close()
    for i in range(0,len(bw_array)):
      #print "for"
      if bw_array[i][1] <= 8:
        diff = diff + (8-bw_array[i][1])
        bw_array[i][1] = 8
      elif diff > 0:
        #print "elif"
        if bw_array[i][1] <= diff:
          res = bw_array[i][1] - 8
          bw_array[i][1] = 8
          diff = diff - res
        else:
          #print "else"
          leftovers = bw_array[i][1] - diff
          if leftovers < 8:
            diff = diff - (bw_array[i][1] - 8)
            bw_array[i][1] = 8
          else:
            bw_array[i][1] = leftovers
            diff = 0
    return bw_array

# throttle the link based on
def throttleLink(bw_a, interface_id, run_duration, **kwargs):
    logging.debug("Throttling the link")
    num_of_sec = 0
    sec_step = 0
    logging.debug("Sess no. " + str(kwargs["num_ses"]))
    for i in range(len(bw_a)):
        print("BW: " + str(bw_a[i][1]) + " sleep: " + str(bw_a[i][0]/1000))
        os.system("tc class change dev %s parent 1:0 classid 1:1 htb rate %fkbit ceil %fkbit"%(interface_id, bw_a[i][1],bw_a[i][1]))
        num_of_sec = num_of_sec + bw_a[i][0]/1000

        if num_of_sec >=  run_duration:
            logging.debug( "Over " + str(run_duration) + "...")
            os.system("tc class change dev %s parent 1:0 classid 1:1 htb rate %fkbit ceil %fkbit"%(interface_id, 100000,100000))
            initilizeLink(10000, 40, 1000, interface_id)
            return
        else:
            sleep(bw_a[i][0]/1000)

    initilizeLink(10000, 40, 1000, interface_id)
    return

# initialise the bandwidth
def initilizeLink(init_bw, delay, queue_size, interface):
    os.system("ethtool -K %s sg off"%interface)
    os.system("ethtool -K %s tso off"%interface)
    os.system("ethtool -K %s gso off"%interface)
    os.system("ethtool -K %s gro off"%interface)
    os.system("bash ./dashc_4K_DP/config_files/tc_cmd.sh %d %d %d %s"%(init_bw, delay-10, queue_size, interface))
    sleep(1)


# get the log file from the client
def copyLogFilesFromClient(client_path, local_folder, ssh_client, client_name):
    print( "client_path: "+ client_path )
    print( "local_folder: "+ local_folder )
    print( "ssh_client: "+ str(ssh_client) )

    #stdin, stdout, stderr = ssh_client.exec_command( 'cd ./Code/dashc-updated; ls -al' )
    #print( stdout.read() )
    #print( stderr.read() )

    # get the log file from the client

    #instead of stfp, lets try to use a simple cp



#    sftp = ssh_client.open_sftp()

    print( "client can connect and open sftp" )

    #dirlist = sftp.listdir(client_path)

    #print( "dirlist: "+str(dirlist) )
    #logging.info( "dirlist: "+str(dirlist) )


    # create the log location
    print( "get fileLoction from: ", local_folder)
    fileLocation = os.path.abspath(local_folder)
    print("got fileLocation = ", fileLocation)

    os.system("mkdir -p %s"%fileLocation)
    print("made directory: ",fileLocation)

    logging.info( "fileLocation: "+fileLocation )
    print("logged: ", fileLocation)

    # get the log file and save it
    crashed_on_first = True
    print ("/vagrant/dashc-updated-algorithms/calculate_qoe.sh " + client_path)
    while (True):
        try:
            os.system("/vagrant/dashc-updated-algorithms/calculate_qoe.sh " + client_path)
            newFileLocation = str(os.path.join(fileLocation+"/logfile.txt"))
            #fileLocation = os.path.abspath(fileLocation)
            logging.debug( newFileLocation )
#                sftp.get(str(os.path.join(client_path,"logfile.txt")),newFileLocation)
            print("Copy logs: " + "cp "+ client_path + "/logfile.txt " + newFileLocation )
            os.system("cp "+ client_path + "/logfile.txt " + newFileLocation )
            break
        except:
            if (crashed_on_first):
                sleep(2)
                print("Waiting for QoE values to be calculated")
                os.system("/vagrant/dashc-updated-algorithms/calculate_qoe.sh " + client_path)
                crashed_on_first = False
            sleep(2)
            continue

    # if there is nothing on the client, print an error
    else:
        print("WARNING: No content to pull")

    #stdin, stdout, stderr = ssh_client.exec_command( 'cd /vagrant/dashc-updated-algorithms' )
    #print( stdout.read() )
    #print( stderr.read() )

    # close the connection
#    sftp.close()

# connect to the client via ssh
def connectToClient(address, user_name, pass_word):
    client = paramiko.SSHClient()
    client.load_system_host_keys()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    client.connect(address, username=user_name, password=pass_word)
    return client

def getValueFromPropertiesFile(full_algo_name, section_name, option_name):
    # root location for client config folder
    root_algo_location=root_location(cfg.client_config_location)
    # read the config file and return the value needed
    parser = configparser.ConfigParser()
    parser.read(root_algo_location+full_algo_name+'.properties')
    return parser.get(section_name, option_name)

# connectCallPlayerOnClient: call the player on the client
# address - IP address of client
# user_name - user name on remote device
# pass_word - password on remote device
# algo_name - algorithm name
# folder_time - folder location for log files on remote device - based on epoch time in seconds
# client_name - client number e.g. Client-1
# client_delay - how long to delay the client before it start => # clients * 2 seconds
def connectCallPlayerOnClient(address, user_name, pass_word, algo_name, folder_time, client_name, client_delay):
    # ssh to client
    client=connectToClient(address, user_name, pass_word)
    # get the value for a given section and option
    full_algo_name=client_name+"-"+algo_name
    movie_clip=getValueFromPropertiesFile(full_algo_name, client_name, '-movie_url')
    adapt=getValueFromPropertiesFile(full_algo_name, 'dashc', '-adapt')
    extended=getValueFromPropertiesFile(full_algo_name, 'dashc', '-extended')
    no_segs=getValueFromPropertiesFile(full_algo_name, 'dashc', '-no-segs')
    max_rep_height=getValueFromPropertiesFile(full_algo_name, 'dashc', '-max-rep-height')

    # we need to define how long the player should sleep before calling dashc
    client_command="cd /vagrant/dashc-updated-algorithms; mkdir -p ./Logs/" + folder_time + "/" + client_name + " ; sleep " + str(client_delay) + "; ./dashc.exe play " + movie_clip + " -turnlogon true -extended "+ extended + " -no-segs " + no_segs +" -max-rep-height " + max_rep_height +" -adapt " + adapt + " -logname ./Logs/" + folder_time + "/" + client_name + " -segmentlist local"
    print ("cd /vagrant/dashc-updated-algorithms; mkdir -p ./Logs/" + folder_time + "/" + client_name + " ; sleep " + str(client_delay) + "; ./dashc.exe play " + movie_clip + " -turnlogon true -extended "+ extended + " -no-segs " + no_segs +" -max-rep-height " + max_rep_height +" -adapt " + adapt + " -logname ./Logs/" + folder_time + "/" + client_name + " -segmentlist local")
    # log and call the client command
    logging.debug( client_command )
    client.exec_command(client_command)

    # I have to keep 'stdin' here or I get an error
    # stdin, stdout, stderr = client.exec_command(client_command)

    # print stdin.read() - don't uncomment as it gives "file error"
    # print stdout.read()
    # print stderr.read()

    #'export DISPLAY=:0; cd ~/Code/dashc-updated; ls -al; ./dashc.exe play http\://192.168.1.2/sintel_4sec/x264/sintel/DASH_Files/full/sintel_enc_x264_dash.mpd; /bin/bash')
    #client.exec_command('mkdir -p /sdcard/Logs_ExoPlayer/')
    #client.exec_command('mkdir -p /sdcard/Logs_Battery/')
    #client.exec_command('rm /sdcard/Logs_ExoPlayer/*')
    #client.exec_command('rm /sdcard/Logs_Battery/*')
    #client.exec_command("dumpsys batterystats --reset")
    return client

# config_parser_output: create the cofig properties file for this client
# algo_name - name of algorithm, e.g. bba-2, conv
# number_segments - number of segments to stream, e.g. 75
# num_client - id number for this client, e.g. 1
# full_algo_name - combination of num_client and algo_name, e.g. Client-1-bba-2
# clip_url - clip url to stream
def config_parser_output( algo_name, number_segments, num_client, full_algo_name, clip_url ):
    # logging details
    logging.info( "create config file for " + full_algo_name )
    logging.debug( "\nconfig_parser_output")
    logging.debug( "algo_name " + algo_name )
    logging.debug( "number_segments " + str(number_segments) )
    logging.debug( "num_client " + str(num_client) )
    logging.debug( "clip_url " + str(clip_url) )
    logging.debug( "full_algo_name " + full_algo_name + "\n" )

    # values used by all
    config = configparser.ConfigParser()
    clip_section="Client-"+str(num_client)
    config.add_section(clip_section)
    config.set(clip_section, '-movie_url', cfg.released[clip_url] )
    config.set(clip_section, '-movie_id', '1' )
    config.set(clip_section, '-Client_number', str(num_client))

    # default settings for our dashc testbed runs
    player_section="dashc"
    config.add_section(player_section)
    config.set(player_section, '-adapt', algo_name )
    config.set(player_section, '-extended', 'true')
    config.set(player_section, '-no-segs', str(int(number_segments)))
    config.set(player_section, '-max-rep-height', '1080')

    # if algo_name == "basic":
    # config values for basic
    #    config.set(algo_name, 'buffer_level', 30000 )
    #    config.set(algo_name, 're_buffer_level', 4000 )

    # root location for this folder
    root_algo_location=root_location(cfg.client_config_location)

    # write the file and change the file propoerties
    with open(root_algo_location+full_algo_name+'.properties', 'w') as configfile:
        config.write(configfile)
        os.system("chmod 777 -R * %s/%s.properties"% (str(root_algo_location), str(full_algo_name)))
        # sleep(1)
        with open(root_algo_location+full_algo_name+'.properties', 'r') as fin:
            data = fin.read().splitlines(True)
        with open(root_algo_location+full_algo_name+'.properties', 'w') as fout:
            fout.writelines(data[1:])
    return

    # Python 3
    # basic
    # config = ConfigParser.ConfigParser()
    # set the clip to stream
    # config[algo_name]['movie_url'] = "http\://192.168.1.2/sintel_4sec/x264/sintel/DASH_Files/full/sintel_enc_x264_dash.mpd"
    # config[algo_name]['movie_id'] = 1
    # config[algo_name]['algorithm_type'] = algo_name

    # if algo_name == "basic":
    #     # config values for basic
    #     config[algo_name]['buffer_level'] = 30000
    #     config[algo_name]['re_buffer_level'] = 4000
    #     config[algo_name]['min_buffer_level'] = 26000
    #     config[algo_name]['playback_buffer_level'] = 8000
    #     config[algo_name]['log_prefix'] = "B_2018.01.17_10.42.37_R1"
    #     config[algo_name]['use_prediction'] = 0
    #     config[algo_name]['options'] = 0
    #     config[algo_name]['segDur'] = 4
    #     config[algo_name]['numSegs'] = 75


def printCommand():
    print("This python script takes 8 parameters:")
    print("sudo python runExperimentsExoMD.py interface_id num_clients num_runs run_duration algorithms debug")
    print("interface_id - interface id for the network connected to the wireless router, e.g., enp3s0")
    print("num_clients - \tnumber of competing clients, e.g., 2")
    print("num_runs - \tnumber of runs for each trace file, e.g., 1")
    print("run_duration - \tnumber of seconds to stream the clip. e.g. 10")
    print("\t\t(30 seconds will be added for collecting files from clients)")
    print("algorithms - \tan array of algorithms to use - no spaces. e.g., [conv,bba-2,arbiter]")
    print("segment size -\tduration of segment in seconds, e.g., 4")
    print("clip_urls - \ta list of clips to use - no spaces. e.g., [sintel_2sec,sintel_4sec,sstb_2sec]")
    print("debug - \toptional - used to increase the number of print statements")


if __name__ == '__main__':

    # if less than 6 parameters passed, then print statement and stop
    if len(sys.argv) < 8:
        printCommand()
        sys.exit(1)
    elif len(sys.argv) < 9:
        logging.basicConfig(stream=sys.stderr, level=logging.INFO)
        logging.info( "Logging level INFO" )
    elif len(sys.argv) == 9 :
        logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
        logging.debug( "Logging level DEBUG" )
    else:
        # if more than 7 parameters passed, then print statement and stop
        logging.info( "Too many parameters" )
        printCommand()
        sys.exit(1)

    interface_id = sys.argv[1]
    num_competing_clients = int(sys.argv[2])
    client_delay=num_competing_clients*2
    num_runs_max_range = int(sys.argv[3])+1
    num_runs = list(range(1, num_runs_max_range))
    run_duration = int(sys.argv[4])+30+client_delay
    algorithms = sys.argv[5]
    # split up the algorithms - possible options include@
    # conventional - conv
    # BBA_0 to BBA_2 - bba-2
    # Arbiter - arbiter
    logging.debug( "Algorithms input" + str(algorithms ) )
    algorithms = algorithms.split('[')
    logging.debug( "Algorithms input" + str(algorithms ) )
    algorithms = algorithms[1].split(']')
    logging.debug( "Algorithms input" + str(algorithms ) )
    algorithms = algorithms[0].split(',')
    logging.debug( "Algorithms input" + str(algorithms ) )
    num_algorithms=len(algorithms)

    logging.debug( "Interface_id: " + interface_id )
    logging.debug( "num_clients: " + str(num_competing_clients) )
    logging.debug( "num_runs: " + str(sys.argv[3]) )
    logging.debug( "run_duration: " + str(run_duration) )
    logging.debug( "num_algorithms: " + str(num_algorithms) )
    logging.debug( algorithms )

    # if the number of clients does not equal the number of algorithms
    if num_algorithms != num_competing_clients:
        print("**The number of clients must equal the number of algorithms in the algorithm list**")
        sys.exit(1)

    segment_duration=int(sys.argv[6])
    number_segments=((run_duration-30)-client_delay)/segment_duration
    logging.debug( "Number of segments= " + str(number_segments) + "\n" )

    clip_urls=sys.argv[7]
    logging.debug( "Clip URLs input" + str(clip_urls ) )
    clip_urls = clip_urls.split('[')
    logging.debug( "Clip URLs input" + str(clip_urls ) )
    clip_urls = clip_urls[1].split(']')
    logging.debug( "Clip URLss input" + str(clip_urls ) )
    clip_urls = clip_urls[0].split(',')
    logging.debug( "Clip URLs input" + str(clip_urls ) )
    num_clips=len(clip_urls)

    # determine how many traces to evaluate
    input_trace_location=root_location(cfg.input_trace_location)
    trace_files = [input_trace_location+f for f in listdir(input_trace_location) if isfile(join(input_trace_location, f))]

    # if the number of clients does not equal the number of clips
    if num_clips != num_competing_clients:
        print("**The number of clients must equal the number of clips in the clip url list**")
        sys.exit(1)

    # begin by emptying the algorithms
    input_algorithms_location=root_location(cfg.client_config_location)
    num_client = 1
    #for algorithm_name in algorithms:
    #    full_algo_name="Client-"+str(num_client)+"-"+algorithm_name
    #    os.system("rm %s/%s.properties"% (str(input_algorithms_location),str(full_algo_name)))
    #    config_parser_output(algorithm_name, number_segments, num_client, full_algo_name, clip_url)
    #    num_client = num_client + 1

    # remove the clients from the previous run
    os.system("rm %s/*.properties"% (str(input_algorithms_location)))
    for i in range(0, num_competing_clients):
        full_algo_name="Client-"+str(i+1)+"-"+algorithms[i]
        config_parser_output(algorithms[i], number_segments, (i+1), full_algo_name, clip_urls[i])


    # determine how many algorithms to evaluate
    alg_config_files = [input_algorithms_location+f for f in listdir(input_algorithms_location) if isfile(join(input_algorithms_location, f))]


    # for the number of runs
    for current_run in num_runs:
        logging.debug( "current run: "+str(current_run) )
        logging.debug( "num_runs: "+str(num_runs) )

        # get time and save as folder name on the client for storing the output


        # determine if 4G or 3G - w
        trace=""

        for trace_file in trace_files:
            delim = ".csv"
            print("Load bw values from trace: " + trace_file)
            if ".csv" in trace_file:
                print("Loading 4G")
                trace="4G-WIFI"
                bw_a = readCsvThr(trace_file)
                # get the prefix for this trace
                prefix_name = trace_file.split("/")[-1].split(".csv")[0] +"_R"\
                        + str(current_run)
            else:
                print("Loading 3G")
                trace="3G"
                delim = ".log"
                bw_a = readLog3G(trace_file)
                # get the prefix for this trace
                prefix_name = trace_file.split("/")[-1].split(".log")[0] +"_R"\
                        + str(current_run)

            logging.debug( "trace prefix name: "+ prefix_name )

            # define run time variables
            list_of_clients = []
            # go through list of algorithms
            #logging.debug( "Number of algorithms: " + str(alg_config_files) )
            #for alg_config_file in alg_config_files:
            folder_time=int(time())
            for i in range(0, num_competing_clients):
                client_name="Client-"+str(i+1)
                # set up the ssh client
                list_of_clients.append(connectCallPlayerOnClient(cfg.IP_range, "vagrant", "vagrant", algorithms[i], str(folder_time), client_name, client_delay))
                sleep(1)

            # initialise the link
            # initial values for "init_bw, delay, queue_size, interface"
            initilizeLink(bw_a[0][1], 40, 1000, interface_id)

            #aribter takes a bit longer to confure, wait for this to finish
            #os.path.isfile("/vagrant/dashc-updated-algorithms/Logs/"+str(folder_time) +  "/" + client_name + "/data-for-qoe-1.temp")
            sleep(1)

            # now throttle the link based on the trace file
            throttleLink(bw_a, interface_id, run_duration, ssh_client = "", num_ses = current_run)
            sleep(1)

            # sleep for a few seconds to make sure all log files are ready for collection
            logging.debug( "sleep for 10 seconds")
            sleep(10)
            logging.debug( "finished sleeping" )

            logging.debug( "\n" )

            print("Experiments done. Copy log files to local machine...")

            # copy the files from the client machine
            #copyLogFilesFromClient("~/Code/dashc-updated/Logs/%s"%str(folder_time), "./dashc_4K_DP/Logs/", list_of_clients[0])
            # don't pass ~ for client folder location
            os.system("sudo tc qdisc del dev enp0s8")
            os.system("tc class show dev enp0s8")
            os.system("chmod 777 -R * ./dashc_4K_DP/Logs/")
            for i in range(0, num_competing_clients):
                client_name="Client-"+str(i+1)
                #os.system("/vagrant/dashc-updated-algorithms/calculate_qoe.sh " + "/vagrant/dashc-updated-algorithms/Logs/" + str(folder_time) + "/" + client_name)
                logging.debug( "geting files from client %d", i+1)
                copyLogFilesFromClient("/vagrant/dashc-updated-algorithms/Logs/"+str(folder_time) +  "/" + client_name, "./dashc_4K_DP/Logs/%s_sec/%s/%s/%s/%s"%(str(segment_duration), trace, prefix_name, str(folder_time), client_name), list_of_clients[i], client_name)
                # close the connection to the client
                list_of_clients[i].close()

            logging.debug( "\n" )

    sys.exit(1)
